﻿#######################
## Sync-VMNames 0.9
##
## Sync-VMNames.ps1
## Last revision date: June 21st, 2013
##
## Author: Eric Siron
## (C) 2013 Altaro Software
##
## A tool to synchronize virtual machine names with their computer names
##

#Region Initialize Environment
#Requires -Version 3

## Script Variables ##
Import-LocalizedData -BindingVariable LocalStrings -FileName Sync-VMNames.psd1
$WindowWidth = $Host.UI.RawUI.WindowSize.Width
$WindowHeight = $Host.UI.RawUI.WindowSize.Height
$VSwitchAvailable = $false		# indicates whether or not a Hyper-V virtual switch can be created/manipulated
$CopyrightStamp = $LocalStrings.Copyright
$BackColor = "Black"
$ForeColor = "Green"
$BackColorHeader = "White"
$ForeColorHeader = "DarkBlue"
$ForeColorExitAndWarnings = "Yellow"
$VMsMismatched = @() # (Index, HypervisorName, GuestName, GuestIP, GuestDomain)
$VMsNoIC = @()
$VMsNotWindows = @()
$VMsNotResponding = @()
$VMsNoIPs = @()
#EndRegion Initialize Environment

#Region UI Functions
function Show-Menu([String]$MenuTitle, [String[]]$ArrayOfOptions, [String]$BackMenuText = $LocalStrings.BackMenu, [String]$AdditionalInfo = "", [System.Management.Automation.SwitchParameter]$NoCancel)
{
	$HeaderString = New-Object System.Text.StringBuilder(6 * $WindowWidth)
	$MenuString = New-Object System.Text.StringBuilder($WindowWidth * $WindowHeight)
	$LineBar = " " + ("-" * ($WindowWidth - 3) + " ")
	$Host.UI.RawUI.WindowTitle = $MenuTitle
	$HeaderString.AppendLine($LineBar) | Out-Null
	$HeaderString.AppendLine("|" + $MenuTitle.PadLeft(($MenuTitle.Length / 2) + ($WindowWidth / 2)).PadRight($WindowWidth - 3) + "|") | Out-Null
	$HeaderString.AppendLine("|" + "".PadLeft($WindowWidth / 2).PadRight($WindowWidth - 3) + "|") | Out-Null
	$HeaderString.AppendLine("|" + $CopyrightStamp.PadLeft(($CopyrightStamp.Length / 2) + ($WindowWidth / 2)).PadRight($WindowWidth - 3) + "|") | Out-Null
	$HeaderString.AppendLine($LineBar) | Out-Null
	$HeaderString.AppendLine() | Out-Null
	if($ArrayOfOptions.Count -gt 0)
	{
		$ArrayOfOptions | ForEach -Begin { $i = 0} -Process {
			$i += 1
			$Line = " [" + $i + "] " + $_
			if($Line.Length -gt ($WindowWidth - 4))
			{ $Line = ($Line.Substring(0, ($WindowWidth - 7)) + "...") }
			$MenuString.AppendLine($Line) | Out-Null
		}
	}
	else
	{ $MenuString.AppendLine($LocalStrings.NoItems) | Out-Null }

	if($AdditionalInfo.Length -gt 0)
	{ $MenuString.AppendLine("`r`n $AdditionalInfo") | Out-Null	}

	Clear-Host
	Write-Host -ForegroundColor $ForeColorHeader -BackgroundColor $BackColorHeader $HeaderString.ToString()
	Write-Host -ForegroundColor $ForeColor -BackgroundColor $BackColor $MenuString.ToString()
	if(-not $NoCancel)
	{ Write-Host -ForegroundColor $ForeColorExitAndWarnings -BackgroundColor $BackColor " [0] $BackMenuText" }
	return (Read-Host -Prompt $LocalStrings.SelectOption)
}
function Show-LoadingVMs()
{
	Clear-Host
	Write-Host $LocalStrings.LoadingVMs
}
function Show-MainMenu()
{
	Show-Menu -MenuTitle $LocalStrings.MainMenuTitle -ArrayOfOptions $LocalStrings.SyncFromGuests, $LocalStrings.SyncFromHypervisor, $LocalStrings.ViewReport -BackMenuText $LocalStrings.Exit
}
function Write-CenteredString
{
	param([String]$StringToDisplay, [ConsoleColor]$ForeColor = "White")
	Write-Host $StringToDisplay.PadLeft(($StringToDisplay.Length / 2) + ($WindowWidth / 2)).PadRight($WindowWidth - 1) -ForegroundColor $ForeColor
}
#EndRegion UI Functions
#Region Worker Functions
function Verify-KVPIC
{
	param([String]$VMName)
	return (Get-VMIntegrationService -Name "Key-Value Pair Exchange" -VMName $VMName).Enabled
}

function Load-Mismatches
{
	$HypervisorNames = @()
	$GuestNames = @()
	$Script:VMsMismatched = @()
	$Script:VMsNoIC = @()
	$Script:VMsNotWindows = @()
	$Script:VMsNotResponding =
	$Script:VMsNoIPs = @()
	Get-VM | ForEach-Object { $HypervisorNames += ,@($_.Name, $_.VMId) }
	$HypervisorNames | ForEach-Object -Begin { $i = 0 } -Process {
		if(-not (Verify-KVPIC -VMName $_[0]))
		{ $Script:VMsNoIC += $_[0] }
		else
		{
			$KVPXMLData = (Get-WmiObject -Namespace root\virtualization\v2 -Query ("Select GuestIntrinsicExchangeItems From Msvm_KvpExchangeComponent Where SystemName = '{0}'" -f $_[1])).GuestIntrinsicExchangeItems
			if($KVPXMLData.Length -eq 0)
			{ $Script:VMsNotResponding += $_[0] }
			else
			{
				# OSName is at position 1
				# Computer name is at position 0
				# Guest IP is at position 17
				# All values are in Property[1]
				if(([xml]$KVPXMLData[1]).Instance.Property[1].Value -notmatch "Windows")
				{ $Script:VMsNotWindows += $_[0] }
				else
				{
					$FullName = ([xml]$KVPXMLData[0]).Instance.Property[1].Value
					$FullName -match "[^\.]*" | Out-Null
					$GuestName = $Matches[0]
					if($GuestName.Length -eq 0)
					{ $Script:VMsNotResponding += $_[0] }
					elseif ($_[0] -ne $GuestName)
					{
						$i++
						$DomainName = ""
						if($GuestName.Length -lt $TestMachine.Length)
						{
							$DomainName = $TestMachine.Substring(($GuestName.Length) + 1, ($TestMachine.Length - $GuestName.Length) - 1)
						}
						if(([xml]$KVPXMLData[17]).Instance.Property[1].Value.Length -gt 0)
						{ $Script:VMsMismatched += ,@($i, $_[0], $GuestName, ([xml]$KVPXMLData[17]).Instance.Property[1].Value, $DomainName) }
						else
						{ $Script:VMsNoIPs += $_[0] }
					}
				}
			}
		}
	}
}
#EndRegion Worker Functions
#Region Operations
function Start-Main()
{
	$Selection = -1
	while ($Selection -ne 0)
	{
		$Selection = Show-MainMenu
		switch($Selection)
		{
			1 {
				Start-SyncFromGuests
			}
			2 {
				Start-SyncFromHypervisor
			}
			3 {
				Start-ReportToScreen
			}
		}
	}
	Clear-Host
}
function Start-SyncFromGuests
{
	Show-LoadingVMs
	Load-Mismatches
	$IgnoreVMs = New-Object System.Collections.ArrayList
	$Selection = -1
	while ($Selection -ne 0)
	{
		$DisplayItems = @()
		$VMsMismatched | ForEach-Object -Process {
			if($_[0] -in $IgnoreVMs)
			{ $DisplayItems += ($LocalStrings.NoChangeDisplay -f $_[1]) }
			else
			{ $DisplayItems += ($LocalStrings.ChangeVMNameDisplay -f $_[1], $_[2]) }
		}
		$Selection = Show-Menu -MenuTitle $LocalStrings.SyncFromGuestsMenuTitle -ArrayOfOptions $DisplayItems -AdditionalInfo $LocalStrings.SyncFromGuestsAdditional
		if($Selection -match "^[\d]+$" -and $Selection -gt 0 )
		{
			if($Selection -notin $IgnoreVMs)
			{ $IgnoreVMs.Add($Selection) }
			else
			{ $IgnoreVMs.Remove($Selection) }
		}
		elseif (([string]$Selection[0]).ToUpper() -eq "G")
		{
			foreach($VMEntry in $VMsMismatched)
			{
				$ClusterGroup = $null
				$HypervisorName = $VMEntry[1]
				$GuestName = $VMEntry[2]
				try
				{
					if($VMEntry[0] -notin $IgnoreVMs)
					{
						Rename-VM -Name $HypervisorName -NewName $GuestName -ErrorAction Stop
						$ClusterGroup = Get-ClusterResource -VMId ((Get-VM -Name $GuestName).VMId) -ErrorAction SilentlyContinue
						if ($ClusterGroup)
						{
							foreach ($Resource in Get-ClusterResource)
							{
								if($Resource.OwnerGroup -eq $Group.Name)
								{
									$Resource.Name = $Resource.Name -replace $GuestName, $NewName
								}
							}
							$ClusterGroup.Name = $ClusterGroup.Name -replace $GuestName, $NewName
						}
					}
				}
				catch
				{
					Write-Error($LocalStrings.RenameVMError -f $VMEntry[1], $VMEntry[2], $_ )
					Read-Host $LocalStrings.KeyContinue
				}
			}
			$Selection = 0
		}
	}
}

function Start-SyncFromHypervisor
{
	Show-LoadingVMs
	Load-Mismatches
	$IgnoreVMs = New-Object System.Collections.ArrayList
	$Selection = -1
	while ($Selection -ne 0)
	{
		$DisplayItems = @()
		$VMsMismatched | ForEach-Object -Process {
			if($_[0] -in $IgnoreVMs)
			{ $DisplayItems += ($LocalStrings.NoChangeDisplay -f $_[2]) }
			else
			{ $DisplayItems += ($LocalStrings.ChangeComputerNameDisplay -f $_[2], $_[1]) }
		}
		$Selection = Show-Menu -MenuTitle $LocalStrings.SyncFromHypervisorMenuTitle -ArrayOfOptions $DisplayItems -AdditionalInfo $LocalStrings.SyncFromHypervisorAdditional
		if($Selection -match "^[\d]+$" -and $Selection -gt 0 )
		{
			if($Selection -notin $IgnoreVMs)
			{ $IgnoreVMs.Add($Selection) }
			else
			{ $IgnoreVMs.Remove($Selection) }
		}
		elseif ([String]($Selection[0]) -match "[A-Za-z]")
		{
			$Selection = ($Selection.Substring(0, 1)).ToUpper()
			if($Selection -eq "G" -or $Selection -eq "R")
			{
				$Modifier = ""
				Clear-Host
				Write-Host $LocalStrings.RenameStart -ForegroundColor $ForeColorExitAndWarnings
				if($Selection -eq "R") { $Modifier -eq " -Restart -Force" }
				foreach ($VM in $VMsMismatched)
				{
					if($VM[0] -notin $IgnoreVMs)
					{
						# the first two aren't strictly necessary but enhance readability
						$NewName = $VM[1]
						$GuestName = $VM[2]
						$GuestIP = $VM[3].ToString().Split(";")
						$GuestDomain = $VM[4]
						$StopLoop = $false
						foreach($IP in $GuestIP)
						{
							if(-not $StopLoop)
							{
								try
								{
									Write-Host ($LocalStrings.AttemptingRename -f $GuestName, $NewName, $IP)
									Invoke-Expression ("Rename-Computer -ComputerName $IP -NewName $NewName{0} -ErrorAction Stop" -f $Modifier)
									Write-Host ($LocalStrings.RenameSuccess -f $GuestName, $NewName)
									$StopLoop = $true
								}
								catch [System.Exception]
								{
									switch -regex ($_.Exception)
									{
										"0x800706BA" { # Unable to connect to target machine
											Write-Host ($LocalStrings.ComputerConnectionError -f $GuestName, $IP)
										}
										"E_ACCESSDENIED" { # Permissions/credentials problem
											$StopLoop = $true # if this doesn't work, trying other IPs won't help
											Write-Host ($LocalStrings.ComputerAccessDenied -f $env:USERNAME, $GuestName)
											try
											{
												if($GuestDomain.Length -gt 0)
												{
													$Credential = Get-Credential -Message ($LocalStrings.DomainCredPrompt -f $GuestName)
													Invoke-Expression ("Rename-Computer -ComputerName $IP -NewName $NewName{0} -DomainCredential $Credential -ErrorAction Stop" -f $Modifier)
													Write-Host ($LocalStrings.RenameSuccess -f $GuestName, $NewName) -ForegroundColor $ForeColor
												}
												else
												{
													$Credential = (Get-Credential -Message $LocalStrings.LocalCredPrompt -f $GuestName)
													Invoke-Expression ("Rename-Computer -ComputerName $IP -NewName $NewName{0} -LocalCredential $Credential -ErrorAction Stop" -f $Modifier)
													Write-Host ($LocalStrings.RenameSuccess -f $GuestName, $NewName) -ForegroundColor $ForeColor
												}
												Write-Host ($LocalStrings.RenameSuccess -f $GuestName, $NewName) -ForegroundColor $ForeColor
											}
											catch {
												Write-Host ($LocalStrings.RenameFail -f $GuestName, $NewName, $_.Exception) -ForegroundColor $ForeColorExitAndWarnings
											}
										}
										default {
											Write-Host ($LocalStrings.RenameFail -f $GuestName, $NewName, $_.Exception) -ForegroundColor $ForeColorExitAndWarnings
										}
									}
								}
							}
						}
					}
				}
				Read-Host $LocalStrings.ReviewResults
				$Selection = 0
			}
		}
	}
}
function Start-ReportToScreen
{
	Show-LoadingVMs
	Load-Mismatches
	Clear-Host
	Write-CenteredString -StringToDisplay $LocalStrings.MismatchTableName -ForeColor Green
	if($VMsMismatched.Count -gt 0)
	{
		Write-Host ($LocalStrings.HyperVNameHeader.PadRight(20, " ") + $LocalStrings.GuestNameHeader.PadRight(20, " ") + $LocalStrings.DomainHeader.PadRight(25, " ") + $LocalStrings.IPHeader) -ForegroundColor Gray
		Write-Host ("-" * ($WindowWidth - 3)) -ForegroundColor Gray
		foreach($VM in $VMsMismatched)
		{
			Write-Host ($VM[1].PadRight(20, " ") + $VM[2].PadRight(20, " ") + $VM[4].PadRight(25, " ")  + $VM[3])
		}
	}
	else
	{
		Write-CenteredString -StringToDisplay ("-" * $LocalStrings.MismatchTableName.Length) -ForeColor Gray
		Write-CenteredString -StringToDisplay $LocalStrings.NoMismatchedVMs
	}

	Write-Host "`r`n"
	Write-CenteredString -StringToDisplay $LocalStrings.NoICTableName -ForeColor Green
	Write-CenteredString -StringToDisplay ("-" * $LocalStrings.NoICTableName.Length) -ForeColor Gray
	if($VMsNoIC.Count -gt 0)
	{
		foreach($VM in $VMsNoIC)
		{ Write-CenteredString -StringToDisplay $VM }
	}
	else { Write-CenteredString -StringToDisplay $LocalStrings.NoVMsWithoutIC }

	Write-Host "`r`n"
	Write-CenteredString -StringToDisplay $LocalStrings.NoWindowsTableName -ForeColor Green
	Write-CenteredString -StringToDisplay ("-" * $LocalStrings.NoWindowsTableName.Length) -ForeColor Gray
	if($VMsNotWindows.Count -gt 0)
	{
		foreach($VM in $VMsNotWindows)
		{ Write-CenteredString -StringToDisplay $VM }
	}
	else { Write-CenteredString -StringToDisplay $LocalStrings.NoNonwindowsVMs }

	Write-Host "`r`n"
	Write-CenteredString -StringToDisplay $LocalStrings.NotRespondingTableName -ForeColor Green
	Write-CenteredString -StringToDisplay ("-" * $LocalStrings.NotRespondingTableName.Length) -ForeColor Gray
	if($VMsNotResponding.Count -gt 0)
	{
		foreach($VM in $VMsNotResponding)
		{ Write-CenteredString -StringToDisplay $VM }
	}
	else { Write-CenteredString -StringToDisplay $LocalStrings.NoVMsNotResponding }

	Write-Host "`r`n"
	Write-CenteredString -StringToDisplay $LocalStrings.NoIPsHeader -ForeColor Green
	Write-CenteredString -StringToDisplay ("-" * $LocalStrings.NoIPsHeader.Length) -ForeColor Gray
	if($VMsNoIPs.Count -gt 0)
	{
		foreach($VM in $VMsNoIPs)
		{ Write-CenteredString -StringToDisplay $VM }
	}
	else { Write-CenteredString -StringToDisplay $LocalStrings.NoIPless }

	Read-Host ("`r`n" + $LocalStrings.KeyContinue)
}
#EndRegion Operations

Start-Main | Out-Null
